﻿using PuppeteerSharp;
using static System.Net.Mime.MediaTypeNames;

namespace StationProbe
{

    internal enum ElementType
    {
        Page,
        Frame
    }

    internal static class ElementConverter
    {
        public static T? As<T>(this object obj) where T: class
        {
            return obj as T;
        }
    }

    internal class ElementCollection
    {
        private readonly Dictionary<ElementType, object> _objects = new();

        public object this[ElementType type] => _objects[type];

        public void AddObject(ElementType element, object obj)
        {
            _objects.Add(element, obj);
        }
    }

    internal abstract class ProbeTask
    {
        private static string _debugFolder = string.Empty;
        protected readonly Database _db;

        protected ScreenshotOptions _snapshotOpts = new ScreenshotOptions() {Type = ScreenshotType.Png };
        public static bool DebugMode { get; set; }


        public ProbeTask(Database db)
        {
            _db = db;
        }

        public static void InitDebugFolder()
        {
            var drives = DriveInfo.GetDrives();
            var largestDriver = drives.OrderByDescending(x => x.TotalFreeSpace).First();
            var debugFolder = largestDriver.Name + "debug";
            if (!Directory.Exists(debugFolder))
            {
                Directory.CreateDirectory(debugFolder);
            }
            var i = 1;
            while (Directory.Exists(Path.Combine(debugFolder, i.ToString("D6"))))
            {
                i++;
            }
            _debugFolder = Path.Combine(debugFolder, i.ToString("D6"));
            Directory.CreateDirectory(_debugFolder);
        }

        public void SaveDebugSnapshot(string name, byte[] imageData)
        {
            if (Directory.Exists(_debugFolder))
            {
                var file = Path.Combine(_debugFolder, name);
                File.WriteAllBytes(file, imageData);
            }
        }

        public abstract Task<ElementCollection> ExecuteAsync(ElementCollection inputs);
    }

    internal class LoginTask: ProbeTask
    {
        private readonly string _loginUrl;

        public LoginTask(Database db, string loginUrl) :base(db)
        {
            _loginUrl = loginUrl;
        }

        public override async Task<ElementCollection> ExecuteAsync(ElementCollection inputs)
        {
            var page = inputs[ElementType.Page].As<IPage>();
            if (page != null)
            {
                await page.GoToAsync(_loginUrl);

                if(DebugMode)
                {
                    var imageData = await page.ScreenshotDataAsync(_snapshotOpts);
                    var fileName = "100000000.jpg";
                    SaveDebugSnapshot(fileName, imageData);
                }

                //username
                var userElement = await page.WaitForSelectorAsync("input[id=\"userName\"]");
                await userElement.FocusAsync();
                await page.Keyboard.TypeAsync("justin.xing");

                if (DebugMode)
                {
                    var imageData = await page.ScreenshotDataAsync(_snapshotOpts);
                    var fileName = "200000000.jpg";
                    SaveDebugSnapshot(fileName, imageData);
                }

                //password
                var passwordElement = await page.WaitForSelectorAsync("input[id=\"password\"]");
                await passwordElement.FocusAsync();
                await page.Keyboard.TypeAsync("vinno@123");

                if (DebugMode)
                {
                    var imageData = await page.ScreenshotDataAsync(_snapshotOpts);
                    var fileName = "300000000.jpg";
                    SaveDebugSnapshot(fileName, imageData);
                }

                //login
                var loginButtonElement = await page.WaitForSelectorAsync("button[b-kvd270uhz9]");
                await loginButtonElement.ClickAsync();
                await page.WaitForNetworkIdleAsync();

                if (DebugMode)
                {
                    var imageData = await page.ScreenshotDataAsync(_snapshotOpts);
                    var fileName = "400000000.jpg";
                    SaveDebugSnapshot(fileName, imageData);
                }
            }
            return inputs;
        }
    }

    internal class MenuTask:ProbeTask
    {
        public MenuTask(Database db) : base(db)
        {
        }

        public override async Task<ElementCollection> ExecuteAsync(ElementCollection inputs)
        {
            Logger.WriteLine("Finding menu...");
            var page = inputs[ElementType.Page].As<IPage>();
            if (page != null)
            {
                //var frame = await page.WaitForFrameAsync("");
                //inputs.AddObject(ElementType.Frame, frame);

                var menuElement = await page.QuerySelectorAsync("a[href=\"feedback/list\"]");
                await menuElement.ClickAsync();
                await page.WaitForNetworkIdleAsync();
                if (DebugMode)
                {
                    var imageData = await page.ScreenshotDataAsync(_snapshotOpts);
                    var fileName = "500000000.jpg";
                    SaveDebugSnapshot(fileName, imageData);
                }
            }
            return inputs;
        }
    }

    internal class FilterTask : ProbeTask
    {
        private readonly DateTime _start;
        private readonly DateTime _end;

        public int PageCount { get; private set; }

        public int ExamCount { get; private set; }

        public FilterTask(Database db, DateTime start, DateTime end):base(db)
        {
            _start = start;
            _end = end;
        }

        public override async Task<ElementCollection> ExecuteAsync(ElementCollection inputs)
        {
            var page = inputs[ElementType.Page].As<IPage>();
            //var frame = inputs[ElementType.Frame].As<IFrame>();
            if(page!= null) //&& frame != null) 
            {
                var filterElements = await page.QuerySelectorAllAsync("div[class=\"me-3 my-1\"]");
                var filterElement1 = filterElements[3];
                var filterElement2 = filterElements[4];
                
                //show filter dialog
                if (DebugMode)
                {
                    var imageData = await page.ScreenshotDataAsync(_snapshotOpts);
                    var fileName = "600000000.jpg";
                    SaveDebugSnapshot(fileName, imageData);
                }

                //input start
                var openElement1 = await filterElement1.WaitForSelectorAsync("div[class=\"multi-select\"]");
                await openElement1.ClickAsync();
                var optionListElement1 = await filterElement1.WaitForSelectorAsync("div[class=\"dropdown-menu shadow\"]");
                var optionElements = await optionListElement1.QuerySelectorAllAsync("div[class=\"dropdown-item\"]");
                var targetElement = optionElements[0];
                await targetElement.ClickAsync();

                if (DebugMode)
                {
                    var imageData = await page.ScreenshotDataAsync(_snapshotOpts);
                    var fileName = "700000000.jpg";
                    SaveDebugSnapshot(fileName, imageData);
                }

                var openlement2 = await filterElement2.WaitForSelectorAsync("div[class=\"multi-select\"]");
                await openlement2.ClickAsync();
                var optionListElement2 = await filterElement2.WaitForSelectorAsync("div[class=\"dropdown-menu shadow\"]");
                optionElements = await optionListElement2.QuerySelectorAllAsync("div[class=\"dropdown-item\"]");
                targetElement = optionElements[0];
                await targetElement.ClickAsync();

                //input end
                if (DebugMode)
                {
                    var imageData = await page.ScreenshotDataAsync(_snapshotOpts);
                    var fileName = "800000000.jpg";
                    SaveDebugSnapshot(fileName, imageData);
                }

                //await frame.WaitForSelectorAsync("input[_bl_13]");
                //await frame.FocusAsync("input[_bl_13]");
                //await page.Keyboard.TypeAsync("justin.xing");

                var searchTableElement = await page.QuerySelectorAsync("div[class=\"table-search\"]");
                var searchButtonElement = await searchTableElement.QuerySelectorAsync("button[class=\"btn btn-primary btn-xs ms-2\"]");
                await searchButtonElement.ClickAsync();
                await page.WaitForNetworkIdleAsync();

                if (DebugMode)
                {
                    var imageData = await page.ScreenshotDataAsync(_snapshotOpts);
                    var fileName = "900000000.jpg";
                    SaveDebugSnapshot(fileName, imageData);
                }

                var pageinationElement = await page.QuerySelectorAsync("div[class=\"table-pagination\"]");
                var examCountElement = await pageinationElement.QuerySelectorAsync("div[class=\"d-none d-sm-inline-block\"]");
                var examCountStr = await page.EvaluateFunctionAsync<string>("(element) => element.textContent", examCountElement);
                var words = examCountStr.Split(' ');
                ExamCount = int.Parse(words[1]);

                var pagesElement = await page.QuerySelectorAsync("ul[class=\"pagination\"]");
                var pageItemElements = await pagesElement.QuerySelectorAllAsync("li");
                var lastPageElement = pageItemElements[pageItemElements.Length - 2];
                var textElement = await lastPageElement.QuerySelectorAsync("a");
                var lastPageStr = await page.EvaluateFunctionAsync<string>("(element) => element.textContent", textElement);
                PageCount = int.Parse(lastPageStr);
            }
            return inputs;
        }
    }

    internal class PageTask:ProbeTask
    {
        private readonly int _page;
        public PageTask(Database db, int page):base(db)
        {
            _page = page;
        }

        private async Task JumpToPageAsync(IPage page)
        {
            while (true)
            {
                var pagesElement = await page.QuerySelectorAsync("ul[class=\"pagination\"]");
                var pageItemElements = await pagesElement.QuerySelectorAllAsync("li");
                var currentPageElemment = await pagesElement.QuerySelectorAsync("li[class=\"page-item active\"]");
                var textElement = await currentPageElemment.QuerySelectorAsync("a");
                var currentPageStr = await page.EvaluateFunctionAsync<string>("(element) => element.textContent", textElement);
                var currentPage = int.Parse(currentPageStr);
                if(currentPage != _page+1)
                {
                    var nextPageElement = pageItemElements[pageItemElements.Length - 1];
                    await nextPageElement.ClickAsync();
                    await page.WaitForNetworkIdleAsync();
                }
                else
                {
                    break;
                }
            }
            
        }

        public override async Task<ElementCollection> ExecuteAsync(ElementCollection inputs)
        {
            var page = inputs[ElementType.Page].As<IPage>();
            //var frame = inputs[ElementType.Frame].As<IFrame>();
            if (page != null) //&& frame != null)
            {
                await JumpToPageAsync(page);
                if (DebugMode)
                {
                    var imageData = await page.ScreenshotDataAsync(_snapshotOpts);
                    var fileName = "9" + _page.ToString("D5") +"000.jpg";
                    SaveDebugSnapshot(fileName, imageData);
                }
            }
            return inputs;
        }
    }

    internal class ExamTask:ProbeTask
    {
        private readonly int _batchTaskId;
        private readonly int _pageIndex;
        private readonly int _examIndex;
        private readonly int _examIndexInPage;

        public ExamTask(Database db, int batchTaskId, int pageIndex, int examIndex, int examIndexInPage): base(db)
        {
            _batchTaskId = batchTaskId;
            _pageIndex = pageIndex;
            _examIndex = examIndex;
            _examIndexInPage = examIndexInPage;
        }

        private async Task JumpToPageAsync(IPage page)
        {
            while (true)
            {
                var pagesElement = await page.QuerySelectorAsync("ul[class=\"pagination\"]");
                var pageItemElements = await pagesElement.QuerySelectorAllAsync("li");
                var currentPageElemment = await pagesElement.QuerySelectorAsync("li[class=\"page-item active\"]");
                var textElement = await currentPageElemment.QuerySelectorAsync("a");
                var currentPageStr = await page.EvaluateFunctionAsync<string>("(element) => element.textContent", textElement);
                var currentPage = int.Parse(currentPageStr);
                if (currentPage != _pageIndex + 1)
                {
                    var nextPageElement = pageItemElements[pageItemElements.Length - 1];
                    await nextPageElement.ClickAsync();
                    await page.WaitForNetworkIdleAsync();
                }
                else
                {
                    break;
                }
            }

        }

        public override async Task<ElementCollection> ExecuteAsync(ElementCollection inputs)
        {
            var page = inputs[ElementType.Page].As<IPage>();
            //var frame = inputs[ElementType.Frame].As<IFrame>();
            if (page != null) //&& frame != null)
            {
                await JumpToPageAsync(page);
                var tableElement = await page.QuerySelectorAsync("table[class=\"table table-bordered table-striped table-hover\"]");
                var examLinkElements = await tableElement.QuerySelectorAllAsync("div[style=\"cursor:pointer;\"]");
                var eaxmLinkElement = examLinkElements[_examIndexInPage];
                var patientName = await page.EvaluateFunctionAsync<string>("(element) => element.textContent", eaxmLinkElement);
                //var examDateStr = await page.EvaluateFunctionAsync<string>("(element) => element.textContent", examLinkElements[6]);
                var examDate = DateTime.Now; //Parse(examDateStr);
                //Enter exam
                await eaxmLinkElement.ClickAsync();
                //await page.WaitForNavigationAsync();
                await page.WaitForNetworkIdleAsync();

                if (DebugMode)
                {
                    var imageData = await page.ScreenshotDataAsync(_snapshotOpts);
                    var fileName = "9" + _pageIndex.ToString("D5") + _examIndexInPage.ToString("D3") +".jpg";
                    SaveDebugSnapshot(fileName, imageData);
                }

                var exmaId = Guid.NewGuid().ToString("N");
                
                //Get patient info
                //var patientName = "";
                var patientSex = "";
                var patientAge = 0;
                //var examDate = DateTime.Now;

                var report = "";
                var reportElements = await page.QuerySelectorAllAsync("div[class=\"col\"]");
                foreach ( var reportElement in reportElements)
                {
                    var strong = await reportElement.QuerySelectorAsync("strong");
                    if(strong != null)
                    {
                        var content = await page.EvaluateFunctionAsync<string>("(element) => element.textContent", strong);
                        report += " " + content;
                    }
                }

                //Get report
                //var report = "";

                //Save images.
                var imageUrls = new List<string>();
                var imageDivElements = await page.QuerySelectorAllAsync("div[class=\"bb-img is-preview\"]");
                if(imageDivElements != null)
                {
                    for(var i=0;i<imageDivElements.Length;i++)
                    {
                        var imageElement = await imageDivElements[i].WaitForSelectorAsync("img");
                        var imageUrl = await imageElement.GetPropertyAsync("src");
                        var url = await imageUrl.JsonValueAsync<string>();
                        imageUrls.Add(url);
                    }
                    await page.ReloadAsync();
                    Parallel.For(0, imageUrls.Count - 1, async(i) =>
                    {
                        var url = imageUrls[i];
                        var response = await page.WaitForResponseAsync(response => (response.Request.Url == url && response.Ok));
                        var buffer = await response.BufferAsync();
                        var image = new Image()
                        {
                            Id = Guid.NewGuid().ToString("N"),
                            ExamId = exmaId,
                            ImageIndex = i,
                            Data = buffer,
                            CreateTime = DateTime.Now,
                        };
                        _db.AddImage(image);
                    });
                }
                //var imageUrls = new List<string>();
                //for (var i = 0; i < imageUrls.Count; i++)
                //{
                //    var response = await page.WaitForResponseAsync(imageUrls[i]);
                //    var buffer = await response.BufferAsync();
                //    var image = new Image()
                //    {
                //        Id = Guid.NewGuid().ToString("N"),
                //        ExamId = exmaId,
                //        ImageIndex = i,
                //        Data = buffer,
                //        CreateTime = DateTime.Now,
                //    };
                //    _db.AddImage(image);
                //}

                //Save exam.
                var exam = new Exam()
                {
                    Id = exmaId,
                    PatientName = patientName,
                    PatientSex = patientSex,
                    PatientAge = patientAge,
                    ExamDate = examDate,
                    Report = report,
                    BatchTaskId = _batchTaskId,
                    ExamIndex = _examIndex,
                    PageIndex = _pageIndex,
                    ExamIndexInPage = _examIndexInPage,
                    CreateTime = DateTime.Now
                };
                _db.AddExam(exam);
                //Back to page.
                await page.GoBackAsync();
                if (DebugMode)
                {
                    var imageData = await page.ScreenshotDataAsync(_snapshotOpts);
                    var fileName = "9" + _pageIndex.ToString("D5") + "999.jpg";
                    SaveDebugSnapshot(fileName, imageData);
                }
            }
            return inputs;
        }
    }
}
